# Generated from ECMAScript.g4 by ANTLR 4.10
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .ECMAScriptParser import ECMAScriptParser
else:
    from ECMAScriptParser import ECMAScriptParser


# This class defines a complete listener for a parse tree produced by ECMAScriptParser.
class ECMAScriptListener(ParseTreeListener):

    # Enter a parse tree produced by ECMAScriptParser#program.
    def enterProgram(self, ctx: ECMAScriptParser.ProgramContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#program.
    def exitProgram(self, ctx: ECMAScriptParser.ProgramContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#sourceElements.
    def enterSourceElements(self, ctx: ECMAScriptParser.SourceElementsContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#sourceElements.
    def exitSourceElements(self, ctx: ECMAScriptParser.SourceElementsContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#sourceElement.
    def enterSourceElement(self, ctx: ECMAScriptParser.SourceElementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#sourceElement.
    def exitSourceElement(self, ctx: ECMAScriptParser.SourceElementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#statement.
    def enterStatement(self, ctx: ECMAScriptParser.StatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#statement.
    def exitStatement(self, ctx: ECMAScriptParser.StatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#block.
    def enterBlock(self, ctx: ECMAScriptParser.BlockContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#block.
    def exitBlock(self, ctx: ECMAScriptParser.BlockContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#statementList.
    def enterStatementList(self, ctx: ECMAScriptParser.StatementListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#statementList.
    def exitStatementList(self, ctx: ECMAScriptParser.StatementListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#variableStatement.
    def enterVariableStatement(self, ctx: ECMAScriptParser.VariableStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#variableStatement.
    def exitVariableStatement(self, ctx: ECMAScriptParser.VariableStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#variableDeclarationList.
    def enterVariableDeclarationList(self, ctx: ECMAScriptParser.VariableDeclarationListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#variableDeclarationList.
    def exitVariableDeclarationList(self, ctx: ECMAScriptParser.VariableDeclarationListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#variableDeclaration.
    def enterVariableDeclaration(self, ctx: ECMAScriptParser.VariableDeclarationContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#variableDeclaration.
    def exitVariableDeclaration(self, ctx: ECMAScriptParser.VariableDeclarationContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#initialiser.
    def enterInitialiser(self, ctx: ECMAScriptParser.InitialiserContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#initialiser.
    def exitInitialiser(self, ctx: ECMAScriptParser.InitialiserContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#emptyStatement_.
    def enterEmptyStatement_(self, ctx: ECMAScriptParser.EmptyStatement_Context):
        pass

    # Exit a parse tree produced by ECMAScriptParser#emptyStatement_.
    def exitEmptyStatement_(self, ctx: ECMAScriptParser.EmptyStatement_Context):
        pass

    # Enter a parse tree produced by ECMAScriptParser#expressionStatement.
    def enterExpressionStatement(self, ctx: ECMAScriptParser.ExpressionStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#expressionStatement.
    def exitExpressionStatement(self, ctx: ECMAScriptParser.ExpressionStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ifStatement.
    def enterIfStatement(self, ctx: ECMAScriptParser.IfStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ifStatement.
    def exitIfStatement(self, ctx: ECMAScriptParser.IfStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#DoStatement.
    def enterDoStatement(self, ctx: ECMAScriptParser.DoStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#DoStatement.
    def exitDoStatement(self, ctx: ECMAScriptParser.DoStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#WhileStatement.
    def enterWhileStatement(self, ctx: ECMAScriptParser.WhileStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#WhileStatement.
    def exitWhileStatement(self, ctx: ECMAScriptParser.WhileStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ForStatement.
    def enterForStatement(self, ctx: ECMAScriptParser.ForStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ForStatement.
    def exitForStatement(self, ctx: ECMAScriptParser.ForStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ForVarStatement.
    def enterForVarStatement(self, ctx: ECMAScriptParser.ForVarStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ForVarStatement.
    def exitForVarStatement(self, ctx: ECMAScriptParser.ForVarStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ForInStatement.
    def enterForInStatement(self, ctx: ECMAScriptParser.ForInStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ForInStatement.
    def exitForInStatement(self, ctx: ECMAScriptParser.ForInStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ForVarInStatement.
    def enterForVarInStatement(self, ctx: ECMAScriptParser.ForVarInStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ForVarInStatement.
    def exitForVarInStatement(self, ctx: ECMAScriptParser.ForVarInStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#continueStatement.
    def enterContinueStatement(self, ctx: ECMAScriptParser.ContinueStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#continueStatement.
    def exitContinueStatement(self, ctx: ECMAScriptParser.ContinueStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#breakStatement.
    def enterBreakStatement(self, ctx: ECMAScriptParser.BreakStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#breakStatement.
    def exitBreakStatement(self, ctx: ECMAScriptParser.BreakStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#returnStatement.
    def enterReturnStatement(self, ctx: ECMAScriptParser.ReturnStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#returnStatement.
    def exitReturnStatement(self, ctx: ECMAScriptParser.ReturnStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#withStatement.
    def enterWithStatement(self, ctx: ECMAScriptParser.WithStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#withStatement.
    def exitWithStatement(self, ctx: ECMAScriptParser.WithStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#switchStatement.
    def enterSwitchStatement(self, ctx: ECMAScriptParser.SwitchStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#switchStatement.
    def exitSwitchStatement(self, ctx: ECMAScriptParser.SwitchStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#caseBlock.
    def enterCaseBlock(self, ctx: ECMAScriptParser.CaseBlockContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#caseBlock.
    def exitCaseBlock(self, ctx: ECMAScriptParser.CaseBlockContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#caseClauses.
    def enterCaseClauses(self, ctx: ECMAScriptParser.CaseClausesContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#caseClauses.
    def exitCaseClauses(self, ctx: ECMAScriptParser.CaseClausesContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#caseClause.
    def enterCaseClause(self, ctx: ECMAScriptParser.CaseClauseContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#caseClause.
    def exitCaseClause(self, ctx: ECMAScriptParser.CaseClauseContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#defaultClause.
    def enterDefaultClause(self, ctx: ECMAScriptParser.DefaultClauseContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#defaultClause.
    def exitDefaultClause(self, ctx: ECMAScriptParser.DefaultClauseContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#labelledStatement.
    def enterLabelledStatement(self, ctx: ECMAScriptParser.LabelledStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#labelledStatement.
    def exitLabelledStatement(self, ctx: ECMAScriptParser.LabelledStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#throwStatement.
    def enterThrowStatement(self, ctx: ECMAScriptParser.ThrowStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#throwStatement.
    def exitThrowStatement(self, ctx: ECMAScriptParser.ThrowStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#tryStatement.
    def enterTryStatement(self, ctx: ECMAScriptParser.TryStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#tryStatement.
    def exitTryStatement(self, ctx: ECMAScriptParser.TryStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#catchProduction.
    def enterCatchProduction(self, ctx: ECMAScriptParser.CatchProductionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#catchProduction.
    def exitCatchProduction(self, ctx: ECMAScriptParser.CatchProductionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#finallyProduction.
    def enterFinallyProduction(self, ctx: ECMAScriptParser.FinallyProductionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#finallyProduction.
    def exitFinallyProduction(self, ctx: ECMAScriptParser.FinallyProductionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#debuggerStatement.
    def enterDebuggerStatement(self, ctx: ECMAScriptParser.DebuggerStatementContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#debuggerStatement.
    def exitDebuggerStatement(self, ctx: ECMAScriptParser.DebuggerStatementContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#functionDeclaration.
    def enterFunctionDeclaration(self, ctx: ECMAScriptParser.FunctionDeclarationContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#functionDeclaration.
    def exitFunctionDeclaration(self, ctx: ECMAScriptParser.FunctionDeclarationContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#formalParameterList.
    def enterFormalParameterList(self, ctx: ECMAScriptParser.FormalParameterListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#formalParameterList.
    def exitFormalParameterList(self, ctx: ECMAScriptParser.FormalParameterListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#functionBody.
    def enterFunctionBody(self, ctx: ECMAScriptParser.FunctionBodyContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#functionBody.
    def exitFunctionBody(self, ctx: ECMAScriptParser.FunctionBodyContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#arrayLiteral.
    def enterArrayLiteral(self, ctx: ECMAScriptParser.ArrayLiteralContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#arrayLiteral.
    def exitArrayLiteral(self, ctx: ECMAScriptParser.ArrayLiteralContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#elementList.
    def enterElementList(self, ctx: ECMAScriptParser.ElementListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#elementList.
    def exitElementList(self, ctx: ECMAScriptParser.ElementListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#elision.
    def enterElision(self, ctx: ECMAScriptParser.ElisionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#elision.
    def exitElision(self, ctx: ECMAScriptParser.ElisionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#objectLiteral.
    def enterObjectLiteral(self, ctx: ECMAScriptParser.ObjectLiteralContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#objectLiteral.
    def exitObjectLiteral(self, ctx: ECMAScriptParser.ObjectLiteralContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#propertyNameAndValueList.
    def enterPropertyNameAndValueList(self, ctx: ECMAScriptParser.PropertyNameAndValueListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#propertyNameAndValueList.
    def exitPropertyNameAndValueList(self, ctx: ECMAScriptParser.PropertyNameAndValueListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PropertyExpressionAssignment.
    def enterPropertyExpressionAssignment(self, ctx: ECMAScriptParser.PropertyExpressionAssignmentContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PropertyExpressionAssignment.
    def exitPropertyExpressionAssignment(self, ctx: ECMAScriptParser.PropertyExpressionAssignmentContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PropertyGetter.
    def enterPropertyGetter(self, ctx: ECMAScriptParser.PropertyGetterContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PropertyGetter.
    def exitPropertyGetter(self, ctx: ECMAScriptParser.PropertyGetterContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PropertySetter.
    def enterPropertySetter(self, ctx: ECMAScriptParser.PropertySetterContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PropertySetter.
    def exitPropertySetter(self, ctx: ECMAScriptParser.PropertySetterContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#propertyName.
    def enterPropertyName(self, ctx: ECMAScriptParser.PropertyNameContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#propertyName.
    def exitPropertyName(self, ctx: ECMAScriptParser.PropertyNameContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#propertySetParameterList.
    def enterPropertySetParameterList(self, ctx: ECMAScriptParser.PropertySetParameterListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#propertySetParameterList.
    def exitPropertySetParameterList(self, ctx: ECMAScriptParser.PropertySetParameterListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#arguments.
    def enterArguments(self, ctx: ECMAScriptParser.ArgumentsContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#arguments.
    def exitArguments(self, ctx: ECMAScriptParser.ArgumentsContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#argumentList.
    def enterArgumentList(self, ctx: ECMAScriptParser.ArgumentListContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#argumentList.
    def exitArgumentList(self, ctx: ECMAScriptParser.ArgumentListContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#expressionSequence.
    def enterExpressionSequence(self, ctx: ECMAScriptParser.ExpressionSequenceContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#expressionSequence.
    def exitExpressionSequence(self, ctx: ECMAScriptParser.ExpressionSequenceContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#TernaryExpression.
    def enterTernaryExpression(self, ctx: ECMAScriptParser.TernaryExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#TernaryExpression.
    def exitTernaryExpression(self, ctx: ECMAScriptParser.TernaryExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#LogicalAndExpression.
    def enterLogicalAndExpression(self, ctx: ECMAScriptParser.LogicalAndExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#LogicalAndExpression.
    def exitLogicalAndExpression(self, ctx: ECMAScriptParser.LogicalAndExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PreIncrementExpression.
    def enterPreIncrementExpression(self, ctx: ECMAScriptParser.PreIncrementExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PreIncrementExpression.
    def exitPreIncrementExpression(self, ctx: ECMAScriptParser.PreIncrementExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ObjectLiteralExpression.
    def enterObjectLiteralExpression(self, ctx: ECMAScriptParser.ObjectLiteralExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ObjectLiteralExpression.
    def exitObjectLiteralExpression(self, ctx: ECMAScriptParser.ObjectLiteralExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#InExpression.
    def enterInExpression(self, ctx: ECMAScriptParser.InExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#InExpression.
    def exitInExpression(self, ctx: ECMAScriptParser.InExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#LogicalOrExpression.
    def enterLogicalOrExpression(self, ctx: ECMAScriptParser.LogicalOrExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#LogicalOrExpression.
    def exitLogicalOrExpression(self, ctx: ECMAScriptParser.LogicalOrExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#NotExpression.
    def enterNotExpression(self, ctx: ECMAScriptParser.NotExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#NotExpression.
    def exitNotExpression(self, ctx: ECMAScriptParser.NotExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PreDecreaseExpression.
    def enterPreDecreaseExpression(self, ctx: ECMAScriptParser.PreDecreaseExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PreDecreaseExpression.
    def exitPreDecreaseExpression(self, ctx: ECMAScriptParser.PreDecreaseExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ArgumentsExpression.
    def enterArgumentsExpression(self, ctx: ECMAScriptParser.ArgumentsExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ArgumentsExpression.
    def exitArgumentsExpression(self, ctx: ECMAScriptParser.ArgumentsExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ThisExpression.
    def enterThisExpression(self, ctx: ECMAScriptParser.ThisExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ThisExpression.
    def exitThisExpression(self, ctx: ECMAScriptParser.ThisExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#FunctionExpression.
    def enterFunctionExpression(self, ctx: ECMAScriptParser.FunctionExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#FunctionExpression.
    def exitFunctionExpression(self, ctx: ECMAScriptParser.FunctionExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#UnaryMinusExpression.
    def enterUnaryMinusExpression(self, ctx: ECMAScriptParser.UnaryMinusExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#UnaryMinusExpression.
    def exitUnaryMinusExpression(self, ctx: ECMAScriptParser.UnaryMinusExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#AssignmentExpression.
    def enterAssignmentExpression(self, ctx: ECMAScriptParser.AssignmentExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#AssignmentExpression.
    def exitAssignmentExpression(self, ctx: ECMAScriptParser.AssignmentExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PostDecreaseExpression.
    def enterPostDecreaseExpression(self, ctx: ECMAScriptParser.PostDecreaseExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PostDecreaseExpression.
    def exitPostDecreaseExpression(self, ctx: ECMAScriptParser.PostDecreaseExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#TypeofExpression.
    def enterTypeofExpression(self, ctx: ECMAScriptParser.TypeofExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#TypeofExpression.
    def exitTypeofExpression(self, ctx: ECMAScriptParser.TypeofExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#InstanceofExpression.
    def enterInstanceofExpression(self, ctx: ECMAScriptParser.InstanceofExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#InstanceofExpression.
    def exitInstanceofExpression(self, ctx: ECMAScriptParser.InstanceofExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#UnaryPlusExpression.
    def enterUnaryPlusExpression(self, ctx: ECMAScriptParser.UnaryPlusExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#UnaryPlusExpression.
    def exitUnaryPlusExpression(self, ctx: ECMAScriptParser.UnaryPlusExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#DeleteExpression.
    def enterDeleteExpression(self, ctx: ECMAScriptParser.DeleteExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#DeleteExpression.
    def exitDeleteExpression(self, ctx: ECMAScriptParser.DeleteExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#EqualityExpression.
    def enterEqualityExpression(self, ctx: ECMAScriptParser.EqualityExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#EqualityExpression.
    def exitEqualityExpression(self, ctx: ECMAScriptParser.EqualityExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#BitXOrExpression.
    def enterBitXOrExpression(self, ctx: ECMAScriptParser.BitXOrExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#BitXOrExpression.
    def exitBitXOrExpression(self, ctx: ECMAScriptParser.BitXOrExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#MultiplicativeExpression.
    def enterMultiplicativeExpression(self, ctx: ECMAScriptParser.MultiplicativeExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#MultiplicativeExpression.
    def exitMultiplicativeExpression(self, ctx: ECMAScriptParser.MultiplicativeExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#BitShiftExpression.
    def enterBitShiftExpression(self, ctx: ECMAScriptParser.BitShiftExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#BitShiftExpression.
    def exitBitShiftExpression(self, ctx: ECMAScriptParser.BitShiftExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ParenthesizedExpression.
    def enterParenthesizedExpression(self, ctx: ECMAScriptParser.ParenthesizedExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ParenthesizedExpression.
    def exitParenthesizedExpression(self, ctx: ECMAScriptParser.ParenthesizedExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#AdditiveExpression.
    def enterAdditiveExpression(self, ctx: ECMAScriptParser.AdditiveExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#AdditiveExpression.
    def exitAdditiveExpression(self, ctx: ECMAScriptParser.AdditiveExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#RelationalExpression.
    def enterRelationalExpression(self, ctx: ECMAScriptParser.RelationalExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#RelationalExpression.
    def exitRelationalExpression(self, ctx: ECMAScriptParser.RelationalExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#PostIncrementExpression.
    def enterPostIncrementExpression(self, ctx: ECMAScriptParser.PostIncrementExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#PostIncrementExpression.
    def exitPostIncrementExpression(self, ctx: ECMAScriptParser.PostIncrementExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#BitNotExpression.
    def enterBitNotExpression(self, ctx: ECMAScriptParser.BitNotExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#BitNotExpression.
    def exitBitNotExpression(self, ctx: ECMAScriptParser.BitNotExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#NewExpression.
    def enterNewExpression(self, ctx: ECMAScriptParser.NewExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#NewExpression.
    def exitNewExpression(self, ctx: ECMAScriptParser.NewExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#LiteralExpression.
    def enterLiteralExpression(self, ctx: ECMAScriptParser.LiteralExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#LiteralExpression.
    def exitLiteralExpression(self, ctx: ECMAScriptParser.LiteralExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#ArrayLiteralExpression.
    def enterArrayLiteralExpression(self, ctx: ECMAScriptParser.ArrayLiteralExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#ArrayLiteralExpression.
    def exitArrayLiteralExpression(self, ctx: ECMAScriptParser.ArrayLiteralExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#MemberDotExpression.
    def enterMemberDotExpression(self, ctx: ECMAScriptParser.MemberDotExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#MemberDotExpression.
    def exitMemberDotExpression(self, ctx: ECMAScriptParser.MemberDotExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#MemberIndexExpression.
    def enterMemberIndexExpression(self, ctx: ECMAScriptParser.MemberIndexExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#MemberIndexExpression.
    def exitMemberIndexExpression(self, ctx: ECMAScriptParser.MemberIndexExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#IdentifierExpression.
    def enterIdentifierExpression(self, ctx: ECMAScriptParser.IdentifierExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#IdentifierExpression.
    def exitIdentifierExpression(self, ctx: ECMAScriptParser.IdentifierExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#BitAndExpression.
    def enterBitAndExpression(self, ctx: ECMAScriptParser.BitAndExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#BitAndExpression.
    def exitBitAndExpression(self, ctx: ECMAScriptParser.BitAndExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#BitOrExpression.
    def enterBitOrExpression(self, ctx: ECMAScriptParser.BitOrExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#BitOrExpression.
    def exitBitOrExpression(self, ctx: ECMAScriptParser.BitOrExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#AssignmentOperatorExpression.
    def enterAssignmentOperatorExpression(self, ctx: ECMAScriptParser.AssignmentOperatorExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#AssignmentOperatorExpression.
    def exitAssignmentOperatorExpression(self, ctx: ECMAScriptParser.AssignmentOperatorExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#VoidExpression.
    def enterVoidExpression(self, ctx: ECMAScriptParser.VoidExpressionContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#VoidExpression.
    def exitVoidExpression(self, ctx: ECMAScriptParser.VoidExpressionContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#assignmentOperator.
    def enterAssignmentOperator(self, ctx: ECMAScriptParser.AssignmentOperatorContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#assignmentOperator.
    def exitAssignmentOperator(self, ctx: ECMAScriptParser.AssignmentOperatorContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#literal.
    def enterLiteral(self, ctx: ECMAScriptParser.LiteralContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#literal.
    def exitLiteral(self, ctx: ECMAScriptParser.LiteralContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#numericLiteral.
    def enterNumericLiteral(self, ctx: ECMAScriptParser.NumericLiteralContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#numericLiteral.
    def exitNumericLiteral(self, ctx: ECMAScriptParser.NumericLiteralContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#identifierName.
    def enterIdentifierName(self, ctx: ECMAScriptParser.IdentifierNameContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#identifierName.
    def exitIdentifierName(self, ctx: ECMAScriptParser.IdentifierNameContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#reservedWord.
    def enterReservedWord(self, ctx: ECMAScriptParser.ReservedWordContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#reservedWord.
    def exitReservedWord(self, ctx: ECMAScriptParser.ReservedWordContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#keyword.
    def enterKeyword(self, ctx: ECMAScriptParser.KeywordContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#keyword.
    def exitKeyword(self, ctx: ECMAScriptParser.KeywordContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#futureReservedWord.
    def enterFutureReservedWord(self, ctx: ECMAScriptParser.FutureReservedWordContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#futureReservedWord.
    def exitFutureReservedWord(self, ctx: ECMAScriptParser.FutureReservedWordContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#getter.
    def enterGetter(self, ctx: ECMAScriptParser.GetterContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#getter.
    def exitGetter(self, ctx: ECMAScriptParser.GetterContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#setter.
    def enterSetter(self, ctx: ECMAScriptParser.SetterContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#setter.
    def exitSetter(self, ctx: ECMAScriptParser.SetterContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#eos.
    def enterEos(self, ctx: ECMAScriptParser.EosContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#eos.
    def exitEos(self, ctx: ECMAScriptParser.EosContext):
        pass

    # Enter a parse tree produced by ECMAScriptParser#eof.
    def enterEof(self, ctx: ECMAScriptParser.EofContext):
        pass

    # Exit a parse tree produced by ECMAScriptParser#eof.
    def exitEof(self, ctx: ECMAScriptParser.EofContext):
        pass


del ECMAScriptParser



from __future__ import annotations

import asyncio
import os
import posixpath
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING

import pkg_resources

from streamflow.core import utils
from streamflow.core.data import DataLocation, DataManager, DataType, LOCAL_LOCATION
from streamflow.core.deployment import Connector
from streamflow.data import remotepath
from streamflow.deployment.connector.base import ConnectorCopyKind
from streamflow.deployment.connector.local import LocalConnector
from streamflow.log_handler import logger

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from typing import Optional, MutableMapping, Set, MutableSequence


async def _copy(src_connector: Optional[Connector],
                src_location: Optional[str],
                src: str,
                dst_connector: Optional[Connector],
                dst_locations: Optional[MutableSequence[str]],
                dst: str,
                writable: False) -> None:
    if isinstance(src_connector, LocalConnector):
        await dst_connector.copy(
            src=src,
            dst=dst,
            locations=dst_locations,
            kind=ConnectorCopyKind.LOCAL_TO_REMOTE,
            read_only=not writable)
    elif isinstance(dst_connector, LocalConnector):
        await src_connector.copy(
            src=src,
            dst=dst,
            locations=[src_location],
            kind=ConnectorCopyKind.REMOTE_TO_LOCAL,
            read_only=not writable)
    else:
        await dst_connector.copy(
            src=src,
            dst=dst,
            locations=dst_locations,
            kind=ConnectorCopyKind.REMOTE_TO_REMOTE,
            source_connector=src_connector,
            source_location=src_location,
            read_only=not writable)


class DefaultDataManager(DataManager):

    def __init__(self, context: StreamFlowContext):
        super().__init__(context)
        self.path_mapper = RemotePathMapper(context)

    async def _transfer_from_location(self,
                                      src_connector: Connector,
                                      src_location: str,
                                      src: str,
                                      dst_connector: Connector,
                                      dst_locations: MutableSequence[str],
                                      dst: str,
                                      writable: bool):
        # Register source path among data locations
        path_processor = os.path if isinstance(src_connector, LocalConnector) is None else posixpath
        src = path_processor.abspath(src)
        # Create destination folder
        await remotepath.mkdir(dst_connector, dst_locations, str(Path(dst).parent))
        # Follow symlink for source path
        src = await remotepath.follow_symlink(src_connector, src_location, src)
        primary_locations = self.path_mapper.get(src, DataType.PRIMARY)
        copy_tasks = []
        remote_locations = []
        data_locations = []
        for dst_location in dst_locations:
            # Check if a primary copy of the source path is already present on the destination location
            found_existing_loc = False
            for primary_loc in primary_locations:
                if primary_loc.location == dst_location:
                    # Wait for the source location to be available on the destination path
                    await primary_loc.available.wait()
                    # If yes, perform a symbolic link if possible
                    if not writable:
                        await remotepath.symlink(dst_connector, dst_location, primary_loc.path, dst)
                        self.path_mapper.create_and_map(
                            location_type=DataType.SYMBOLIC_LINK,
                            src_path=src,
                            dst_path=dst,
                            dst_deployment=dst_connector.deployment_name,
                            dst_location=dst_location,
                            available=True)
                    # Otherwise, perform a copy operation
                    else:
                        copy_tasks.append(asyncio.create_task(_copy(
                            src_connector=dst_connector,
                            src_location=dst_location,
                            src=primary_loc.path,
                            dst_connector=dst_connector,
                            dst_locations=[dst_location],
                            dst=dst,
                            writable=True)))
                        data_locations.append(self.path_mapper.put(
                            path=dst,
                            data_location=DataLocation(
                                path=dst,
                                relpath=list(self.path_mapper.get(src))[0].relpath,
                                deployment=dst_connector.deployment_name,
                                data_type=DataType.PRIMARY,
                                location=dst_location,
                                available=False)))
                    found_existing_loc = True
                    break
            # Otherwise, perform a remote copy and mark the destination as primary
            if not found_existing_loc:
                remote_locations.append(dst_location)
                if writable:
                    data_locations.append(self.path_mapper.put(
                        path=dst,
                        data_location=DataLocation(
                            path=dst,
                            relpath=list(self.path_mapper.get(src))[0].relpath,
                            deployment=dst_connector.deployment_name,
                            data_type=DataType.PRIMARY,
                            location=dst_location,
                            available=False)))
                else:
                    data_locations.append(self.path_mapper.create_and_map(
                        location_type=DataType.PRIMARY,
                        src_path=src,
                        dst_path=dst,
                        dst_deployment=dst_connector.deployment_name,
                        dst_location=dst_location))
        # Perform all the copy operations
        if remote_locations:
            copy_tasks.append(asyncio.create_task(_copy(
                src_connector=src_connector,
                src_location=src_location,
                src=src,
                dst_connector=dst_connector,
                dst_locations=remote_locations,
                dst=dst,
                writable=writable)))
        await asyncio.gather(*copy_tasks)
        # Mark all destination data locations as available
        for data_location in data_locations:
            data_location.available.set()

    async def close(self):
        pass

    def get_data_locations(self,
                           path: str,
                           deployment: Optional[str] = None,
                           location_type: Optional[DataType] = None) -> Set[DataLocation]:
        data_locations = self.path_mapper.get(path, location_type)
        data_locations = {loc for loc in data_locations if loc.data_type != DataType.INVALID}
        if deployment is not None:
            data_locations = {loc for loc in data_locations if loc.deployment == deployment}
        return data_locations

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'data_manager.json'))

    def get_source_location(self,
                            path: str,
                            dst_deployment: str) -> Optional[DataLocation]:
        if data_locations := self.get_data_locations(path):
            dst_connector = self.context.deployment_manager.get_connector(dst_deployment)
            same_connector_locations = {loc for loc in data_locations if
                                        loc.deployment == dst_connector.deployment_name}
            if same_connector_locations:
                for loc in same_connector_locations:
                    if loc.data_type == DataType.PRIMARY:
                        return loc
                return list(same_connector_locations)[0]
            else:
                local_locations = {loc for loc in data_locations if isinstance(
                                   self.context.deployment_manager.get_connector(loc.deployment), LocalConnector)}
                if local_locations:
                    for loc in local_locations:
                        if loc.data_type == DataType.PRIMARY:
                            return loc
                    return list(local_locations)[0]
                else:
                    for loc in data_locations:
                        if loc.data_type == DataType.PRIMARY:
                            return loc
                    return list(data_locations)[0]
        else:
            return None

    def invalidate_location(self, location: str, path: str) -> None:
        self.path_mapper.invalidate_location(location, path)

    def register_path(self,
                      deployment: str,
                      location: str,
                      path: str,
                      relpath: Optional[str] = None,
                      data_type: DataType = DataType.PRIMARY) -> DataLocation:
        data_location = DataLocation(
            path=path,
            relpath=relpath or path,
            deployment=deployment,
            location=location or LOCAL_LOCATION,
            data_type=data_type,
            available=True)
        self.path_mapper.put(
            path=path,
            data_location=data_location,
            recursive=True)
        self.context.checkpoint_manager.register(data_location)
        return data_location

    def register_relation(self,
                          src_location: DataLocation,
                          dst_location: DataLocation) -> None:
        data_locations = self.path_mapper.get(src_location.path)
        for data_location in list(data_locations):
            self.path_mapper.put(data_location.path, dst_location)
            self.path_mapper.put(dst_location.path, data_location)

    async def transfer_data(self,
                            src_deployment: str,
                            src_locations: MutableSequence[str],
                            src_path: str,
                            dst_deployment: str,
                            dst_locations: MutableSequence[str],
                            dst_path: str,
                            writable: bool = False):
        # Get connectors and locations from steps
        src_connector = self.context.deployment_manager.get_connector(src_deployment)
        dst_connector = self.context.deployment_manager.get_connector(dst_deployment)
        src_found = False
        for src_location in src_locations:
            if await remotepath.exists(src_connector, src_location, src_path):
                src_found = True
                await self._transfer_from_location(
                    src_connector, src_location, src_path,
                    dst_connector, dst_locations, dst_path,
                    writable)
        # If source path does not exist
        if not src_found:
            # Search it on destination locations
            for dst_location in dst_locations:
                # If it exists, switch source connector and locations with the new ones
                if await remotepath.exists(dst_connector, dst_location, src_path):
                    logger.debug("Path {path} found {location}.".format(
                        path=src_path,
                        location="on location {location}".format(
                            location=dst_location) if dst_location is not None else "on local file-system"
                    ))
                    await self._transfer_from_location(
                        src_connector, dst_location, src_path,
                        dst_connector, dst_locations, dst_path,
                        writable)
                    break


class RemotePathNode(object):
    __slots__ = ('children', 'locations')

    def __init__(self):
        self.children: MutableMapping[str, RemotePathNode] = {}
        self.locations: Set[DataLocation] = set()


class RemotePathMapper(object):

    def __init__(self,
                 context: StreamFlowContext):
        self._filesystem: RemotePathNode = RemotePathNode()
        self.context: StreamFlowContext = context

    def _remove_node(self, location: DataLocation, node: RemotePathNode):
        node.locations.remove(location)
        for n in node.children.values():
            self._remove_node(location, n)

    def create_and_map(self,
                       location_type: DataType,
                       src_path: str,
                       dst_path: str,
                       dst_deployment: str,
                       dst_location: Optional[str],
                       available: bool = False) -> DataLocation:
        data_locations = self.get(src_path)
        dst_data_location = DataLocation(
            path=dst_path,
            relpath=list(data_locations)[0].relpath,
            deployment=dst_deployment,
            data_type=location_type,
            location=dst_location,
            available=available)
        for data_location in list(data_locations):
            self.put(data_location.path, dst_data_location)
            self.put(dst_path, data_location)
        self.put(dst_path, dst_data_location)
        return dst_data_location

    def get(self, path: str, location_type: Optional[DataType] = None) -> Set[DataLocation]:
        path = PurePosixPath(Path(path).as_posix())
        node = self._filesystem
        for token in path.parts:
            if token in node.children:
                node = node.children[token]
            else:
                return set()
        return ({loc for loc in node.locations if loc.data_type == location_type} if location_type
                else node.locations)

    def invalidate_location(self, location: str, path: str) -> None:
        path = PurePosixPath(Path(path).as_posix())
        node = self._filesystem
        for token in path.parts:
            node = node.children[token]
        for loc in node.locations:
            if loc.location == location:
                loc.data_type = DataType.INVALID

    def put(self, path: str, data_location: DataLocation, recursive: bool = False) -> DataLocation:
        path = PurePosixPath(Path(path).as_posix())
        path_processor = utils.get_path_processor(
            self.context.deployment_manager.get_connector(data_location.deployment))
        node = self._filesystem
        nodes = {}
        # Create or navigate hierarchy
        for i, token in enumerate(path.parts):
            node = node.children.setdefault(token, RemotePathNode())
            if recursive:
                nodes[path_processor.join(*path.parts[:i + 1])] = node
        if not recursive:
            nodes[str(path)] = node
        # Process hierarchy bottom-up to add parent locations
        relpath = data_location.relpath
        for node_path in reversed(nodes):
            node = nodes[node_path]
            if node_path == str(path):
                location = data_location
            else:
                location = DataLocation(
                    path=node_path,
                    relpath=relpath if relpath and node_path.endswith(relpath) else path_processor.basename(node_path),
                    deployment=data_location.deployment,
                    data_type=DataType.PRIMARY,
                    location=data_location.location,
                    available=True)
            if location in node.locations:
                break
            else:
                node.locations.add(location)
                relpath = path_processor.dirname(relpath)
        # Return location
        return data_location

    def remove_location(self, location: str):
        data_location = None
        for loc in self._filesystem.locations:
            if loc.location == location:
                data_location = loc
                break
        if data_location:
            self._remove_node(data_location, self._filesystem)



import asyncio.subprocess
from typing import Any, Coroutine, Optional

from streamflow.core.data import StreamWrapper, StreamWrapperContext


class BaseStreamWrapper(StreamWrapper):

    async def close(self):
        if self.closed:
            return
        self.closed = True
        await self.stream.close()

    async def read(self, size: Optional[int] = None):
        return await self.stream.read(size)

    async def write(self, data: Any):
        return await self.stream.write(data)


class StreamReaderWrapper(StreamWrapper):

    async def close(self):
        pass

    async def read(self, size: Optional[int] = None):
        return await self.stream.read(size)

    async def write(self, data: Any):
        raise NotImplementedError


class StreamWriterWrapper(StreamWrapper):

    async def close(self):
        self.stream.close()
        await self.stream.wait_closed()

    async def read(self, size: Optional[int] = None):
        raise NotImplementedError

    async def write(self, data: Any):
        self.stream.write(data)
        await self.stream.drain()


class SubprocessStreamReaderWrapperContext(StreamWrapperContext):

    def __init__(self,
                 coro: Coroutine):
        self.coro: Coroutine = coro
        self.proc: Optional[asyncio.subprocess.Process] = None
        self.stream: Optional[StreamReaderWrapper] = None

    async def __aenter__(self):
        self.proc = await self.coro
        self.stream = StreamReaderWrapper(self.proc.stdout)
        return self.stream

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.proc.wait()
        if self.stream:
            await self.stream.close()


class SubprocessStreamWriterWrapperContext(StreamWrapperContext):

    def __init__(self,
                 coro: Coroutine):
        self.coro: Coroutine = coro
        self.stream: Optional[StreamReaderWrapper] = None

    async def __aenter__(self):
        proc = await self.coro
        self.stream: StreamReaderWrapper(proc.stdout)
        return self.stream

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.stream:
            await self.stream.close()



from __future__ import annotations

import asyncio
import base64
import importlib
import itertools
import os
import posixpath
import uuid
from pathlib import Path
from typing import Any, MutableMapping, MutableSequence, Optional, Set, TYPE_CHECKING, Type, Union

from importlib_metadata import entry_points
from jsonref import loads

from streamflow.core.data import LOCAL_LOCATION
from streamflow.core.exception import InvalidPluginException, WorkflowExecutionException
from streamflow.core.workflow import Token
from streamflow.data import aiotarstream
from streamflow.ext import StreamFlowPlugin
from streamflow.log_handler import logger
from streamflow.workflow.token import IterationTerminationToken, ListToken, ObjectToken, TerminationToken

if TYPE_CHECKING:
    from streamflow.core.config import SchemaEntity
    from streamflow.core.deployment import Connector
    from typing import Iterable


class NamesStack(object):

    def __init__(self):
        self.stack: MutableSequence[Set] = [set()]

    def add_scope(self):
        self.stack.append(set())

    def add_name(self, name: str):
        self.stack[-1].add(name)

    def delete_scope(self):
        self.stack.pop()

    def delete_name(self, name: str):
        self.stack[-1].remove(name)

    def global_names(self) -> Set[str]:
        names = self.stack[0].copy()
        if len(self.stack) > 1:
            for scope in self.stack[1:]:
                names = names.difference(scope)
        return names

    def __contains__(self, name: str) -> bool:
        for scope in self.stack:
            if name in scope:
                return True
        return False


def check_iteration_termination(inputs: Union[Token, Iterable[Token]]) -> bool:
    return check_token_class(inputs, IterationTerminationToken)


def check_termination(inputs: Union[Token, Iterable[Token]]) -> bool:
    return check_token_class(inputs, TerminationToken)


def check_token_class(inputs: Union[Token, Iterable[Token]], cls: Type[Token]):
    if isinstance(inputs, Token):
        return isinstance(inputs, cls)
    else:
        for token in inputs:
            if isinstance(token, MutableSequence):
                if check_token_class(token, cls):
                    return True
            elif isinstance(token, cls):
                return True
        return False


def create_command(command: MutableSequence[str],
                   environment: MutableMapping[str, str] = None,
                   workdir: Optional[str] = None,
                   stdin: Optional[Union[int, str]] = None,
                   stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                   stderr: Union[int, str] = asyncio.subprocess.STDOUT) -> str:
    command = "".join(
        "{workdir}"
        "{environment}"
        "{command}"
        "{stdin}"
        "{stdout}"
        "{stderr}"
    ).format(
        workdir="cd {workdir} && ".format(workdir=workdir) if workdir is not None else "",
        environment="".join(["export %s=\"%s\" && " % (key, value) for (key, value) in
                             environment.items()]) if environment is not None else "",
        command=" ".join(command),
        stdin=" < {stdin}".format(stdin=stdin) if stdin is not None else "",
        stdout=" > {stdout}".format(stdout=stdout) if stdout != asyncio.subprocess.STDOUT else "",
        stderr=(" 2>&1" if stderr == stdout else
                " 2>{stderr}".format(stderr=stderr) if stderr != asyncio.subprocess.STDOUT else
                ""))
    return command


def dict_product(**kwargs) -> MutableMapping[Any, Any]:
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, list(instance)))


async def extract_tar_stream(tar: aiotarstream.AioTarStream,
                             src: str,
                             dst: str,
                             transferBufferSize: Optional[int] = None) -> None:
    async for member in tar:
        if os.path.isdir(dst):
            if posixpath.join('/', member.path) == src:
                member.path = posixpath.basename(member.path)
                await tar.extract(member, dst)
                if member.isdir():
                    dst = os.path.join(dst, member.path)
            else:
                member.path = posixpath.relpath(posixpath.join('/', member.path), src)
                await tar.extract(member, dst)
        elif member.isfile():
            async with await tar.extractfile(member) as inputfile:
                with open(dst, 'wb') as outputfile:
                    while content := await inputfile.read(transferBufferSize):
                        outputfile.write(content)
        else:
            parent_dir = str(Path(dst).parent)
            member.path = posixpath.basename(member.path)
            await tar.extract(member, parent_dir)


def encode_command(command: str):
    return "echo {command} | base64 -d | sh".format(
        command=base64.b64encode(command.encode('utf-8')).decode('utf-8'))


def flatten_list(hierarchical_list):
    if not hierarchical_list:
        return hierarchical_list
    flat_list = []
    for el in hierarchical_list:
        if isinstance(el, MutableSequence):
            flat_list.extend(flatten_list(el))
        else:
            flat_list.append(el)
    return flat_list


def get_class_fullname(cls: Type):
    return cls.__module__ + '.' + cls.__qualname__


def get_class_from_name(name: str) -> Type:
    module_name, class_name = name.rsplit('.', 1)
    return getattr(importlib.import_module(module_name), class_name)


def get_path_processor(connector: Connector):
    return posixpath if connector is not None and connector.deployment_name != LOCAL_LOCATION else os.path


async def get_remote_to_remote_write_command(src_connector: Connector,
                                             src_location: str,
                                             src: str,
                                             dst_connector: Connector,
                                             dst_locations: MutableSequence[str],
                                             dst: str) -> MutableSequence[str]:
    if posixpath.basename(src) != posixpath.basename(dst):
        result, status = await src_connector.run(
            location=src_location,
            command=["test -d \"{path}\"".format(path=src)],
            capture_output=True)
        if status > 1:
            raise WorkflowExecutionException(result)
        # If is a directory
        elif status == 0:
            await asyncio.gather(*(asyncio.create_task(
                dst_connector.run(location=dst_location, command=["mkdir", "-p", dst])
            ) for dst_location in dst_locations))
            return ["tar", "xf", "-", "-C", dst, "--strip-components", "1"]
        # If is a file
        else:
            return ["tar", "xf", "-", "-O", ">", dst]
    else:
        return ["tar", "xf", "-", "-C", posixpath.dirname(dst)]


def get_size(path):
    if os.path.isfile(path):
        return os.path.getsize(path)
    else:
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(path, followlinks=True):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)
        return total_size


def get_tag(tokens: Iterable[Token]) -> str:
    output_tag = '0'
    for tag in [t.tag for t in tokens]:
        if len(tag) > len(output_tag):
            output_tag = tag
    return output_tag


def get_token_value(token: Token) -> Any:
    if isinstance(token, ListToken):
        return [get_token_value(t) for t in token.value]
    elif isinstance(token, ObjectToken):
        return {k: get_token_value(v) for k, v in token.value.items()}
    elif isinstance(token.value, Token):
        return get_token_value(token.value)
    else:
        return token.value


def inject_schema(schema: MutableMapping[str, Any],
                  classes: MutableMapping[str, Type[SchemaEntity]],
                  definition_name: str):
    for name, entity in classes.items():
        if entity_schema := entity.get_schema():
            with open(entity_schema, "r") as f:
                entity_schema = loads(
                    f.read(),
                    base_uri='file://{}/'.format(os.path.dirname(entity_schema)),
                    jsonschema=True)
            schema['definitions'][definition_name]['properties']['type'].setdefault('enum', []).append(name)
            schema['definitions'][definition_name]['definitions'][name] = entity_schema
            schema['definitions'][definition_name].setdefault('allOf', []).append({
                'if': {'properties': {'type': {'const': name}}},
                'then': {'properties': {'config': entity_schema}}})


def load_extensions():
    plugins = entry_points(group='unito.streamflow.plugin')
    for plugin in plugins:
        plugin = (plugin.load())()
        if isinstance(plugin, StreamFlowPlugin):
            plugin.register()
            logger.info("Succesfully registered plugin {}".format(
                get_class_fullname(type(plugin))))
        else:
            raise InvalidPluginException("StreamFlow plugins must extend the streamflow.ext.StreamFlowPlugin class")


def random_name() -> str:
    return str(uuid.uuid4())


def wrap_command(command: str):
    return ["/bin/sh", "-c", "{command}".format(command=command)]



import argparse
import asyncio
import logging
import os
import sys
import uuid

import streamflow.cwl.main
from streamflow.config.config import WorkflowConfig
from streamflow.config.validator import SfValidator
from streamflow.core import utils
from streamflow.core.exception import WorkflowDefinitionException, WorkflowException
from streamflow.log_handler import logger
from streamflow.main import build_context

parser = argparse.ArgumentParser(description='cwl-runner interface')
parser.add_argument('processfile', nargs='?', type=str,
                    help='The CommandLineTool, ExpressionTool, or Workflow description to run. '
                         'Optional if the jobfile has a `cwl:tool` field to indicate which process description to run.')
parser.add_argument('jobfile', nargs='?', type=str,
                    help='The input job document')
parser.add_argument('--name', default=str(uuid.uuid4()), type=str,
                    help='Name of the current workflow. Will be used for search and indexing.')
parser.add_argument('--outdir', default=os.getcwd(), type=str,
                    help='Output directory, defaults to the current directory')
parser.add_argument('--debug', action='store_true', help='Debug-level diagnostic output')
parser.add_argument('--quiet', action='store_true', help='No diagnostic output')
parser.add_argument('--version', action='store_true',
                    help='Report the name and version, then quit without further processing')
parser.add_argument('--streamflow-file', type=str,
                    help='The path to a StreamFlow file specifying deployments and bindings for the workflow steps.')


async def _async_main(args: argparse.Namespace):
    validator = SfValidator()
    config_dir = os.getcwd()
    if args.streamflow_file:
        utils.load_extensions()
        with open(args.streamflow_file) as f:
            streamflow_config = validator.yaml.load(f)
        config_dir = os.path.dirname(args.streamflow_file)
        workflows = streamflow_config.get('workflows', {})
        if len(workflows) == 1:
            workflow_name = list(workflows.keys())[0]
        elif len(workflows) == 0:
            workflow_name = 'cwl-workflow'
            streamflow_config['workflows'][workflow_name] = {}
        else:
            raise WorkflowDefinitionException(
                "A StreamFlow file must contain only one workflow definition when used with cwl-runner.")
    else:
        workflow_name = 'cwl-workflow'
        streamflow_config = {
            'version': 'v1.0',
            'workflows': {
                workflow_name: {}
            }
        }
    streamflow_config['workflows'][workflow_name]['type'] = 'cwl'
    streamflow_config['workflows'][workflow_name]['config'] = {
        'file': args.processfile
    }
    if args.jobfile:
        streamflow_config['workflows'][workflow_name]['config']['settings'] = args.jobfile
    validator.validate(streamflow_config)
    workflow_config = WorkflowConfig(workflow_name, streamflow_config)
    context = build_context(config_dir, streamflow_config, args.outdir)
    try:
        await streamflow.cwl.main.main(
            workflow_config=workflow_config,
            context=context,
            args=args)
    except BaseException as e:
        logger.exception(e)
        sys.exit(1)
    finally:
        await context.close()


def main(args):
    args = parser.parse_args(args)
    if args.version:
        from streamflow.version import VERSION
        print("StreamFlow version {version}".format(version=VERSION))
        return
    if args.quiet:
        logger.setLevel(logging.WARN)
    elif args.debug:
        logger.setLevel(logging.DEBUG)
    asyncio.run(_async_main(args))


def run():
    main(sys.argv[1:])


if __name__ == "__main__":
    main(sys.argv[1:])



from streamflow.recovery.checkpoint_manager import DefaultCheckpointManager, DummyCheckpointManager
from streamflow.recovery.failure_manager import DefaultFailureManager, DummyFailureManager

checkpoint_manager_classes = {
    'default': DefaultCheckpointManager,
    'dummy': DummyCheckpointManager
}

failure_manager_classes = {
    'default': DefaultFailureManager,
    'dummy': DummyFailureManager
}



from __future__ import annotations

import asyncio
import os
from asyncio import Condition
from typing import MutableMapping, Optional, cast

import pkg_resources

from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import Connector
from streamflow.core.exception import FailureHandlingException, UnrecoverableTokenException
from streamflow.core.recovery import FailureManager, ReplayRequest, ReplayResponse
from streamflow.core.workflow import CommandOutput, Job, Status, Step
from streamflow.data import remotepath
from streamflow.log_handler import logger
from streamflow.recovery.recovery import JobVersion
from streamflow.workflow.step import ExecuteStep


async def _cleanup_dir(connector: Connector,
                       location: str,
                       directory: str) -> None:
    await remotepath.rm(connector, location, await remotepath.listdir(connector, location, directory))


class DefaultFailureManager(FailureManager):

    def __init__(self,
                 context: StreamFlowContext,
                 max_retries: Optional[int] = None,
                 retry_delay: Optional[int] = None):
        super().__init__(context)
        self.jobs: MutableMapping[str, JobVersion] = {}
        self.max_retries: int = max_retries
        self.replay_cache: MutableMapping[str, ReplayResponse] = {}
        self.retry_delay: Optional[int] = retry_delay
        self.wait_queues: MutableMapping[str, Condition] = {}

    async def _do_handle_failure(self, job: Job, step: Step) -> CommandOutput:
        # Delay rescheduling to manage temporary failures (e.g. connection lost)
        if self.retry_delay is not None:
            await asyncio.sleep(self.retry_delay)
        if job.name not in self.jobs:
            self.jobs[job.name] = JobVersion(
                job=Job(
                    name=job.name,
                    inputs=dict(job.inputs),
                    input_directory=job.input_directory,
                    output_directory=job.output_directory,
                    tmp_directory=job.tmp_directory),
                outputs=None,
                step=step,
                version=1)
        command_output = await self._replay_job(self.jobs[job.name])
        return command_output

    async def _replay_job(self, job_version: JobVersion) -> CommandOutput:
        job = job_version.job
        # Retry job execution until the max number of retries is reached
        if self.max_retries is None or self.jobs[job.name].version < self.max_retries:
            # Update version
            self.jobs[job.name].version += 1
            try:
                # Manage job rescheduling
                allocation = self.context.scheduler.get_allocation(job.name)
                connector = self.context.scheduler.get_connector(job.name)
                locations = self.context.scheduler.get_locations(job.name)
                available_locations = await connector.get_available_locations(
                    service=allocation.target.service)
                active_locations = self.context.scheduler.get_locations(job.name, [Status.RUNNING])
                # If there are active locations, the job just failed
                if active_locations:
                    # If some locations are dead
                    if not all(res in available_locations for res in active_locations):
                        # Notify job failure
                        await self.context.scheduler.notify_status(job.name, Status.FAILED)
                        # Invalidate locations
                        for location in (set(active_locations) - set(available_locations)):
                            self.context.data_manager.invalidate_location(location, '/')
                        # TODO
                    # Otherwise, empty output and tmp folders and re-execute the job
                    cleanup_tasks = []
                    for location in locations:
                        for directory in [job.output_directory, job.tmp_directory]:
                            cleanup_tasks.append(asyncio.create_task(_cleanup_dir(connector, location, directory)))
                            self.context.data_manager.invalidate_location(location, directory)
                    await asyncio.gather(*cleanup_tasks)
                    # TODO: try to do this in a more general way
                    return await cast(ExecuteStep, job_version.step).command.execute(job)
                # Otherwise, the job already completed but must be rescheduled
                else:
                    # TODO
                    # Recover input tokens
                    recovered_tokens = []
                    for token_name, token in job.inputs.items():
                        token_processor = job.step.input_token_processors[token_name]
                        version = 0
                        while True:
                            try:
                                recovered_tokens.append(await token_processor.recover_token(job, locations, token))
                                break
                            except UnrecoverableTokenException as e:
                                version += 1
                                reply_response = await self.replay_job(
                                    ReplayRequest(job.name, e.token.job, version))
                                input_port = job.step.workflow.ports[job.step.input_ports[token.name]]
                                recovered_token = reply_response.outputs[input_port.name]
                                token = await _replace_token(job, token_processor, token, recovered_token)
                                token.name = input_port.name
                    job.inputs = recovered_tokens
            # When receiving a FailureHandlingException, simply fail
            except FailureHandlingException as e:
                logger.exception(e)
                raise
            # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
            except KeyboardInterrupt:
                raise
            except BaseException as e:
                logger.exception(e)
                return await self.handle_exception(job, job_version.step, e)
        else:
            logger.error("Job {name} failed {version} times. Execution aborted".format(
                name=job.name, version=self.jobs[job.name].version))
            raise FailureHandlingException()

    async def close(self):
        pass

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'default_failure_manager.json'))

    async def handle_exception(self, job: Job, step: Step, exception: BaseException) -> CommandOutput:
        logger.info("Handling {exception} failure for job {job}".format(
            job=job.name, exception=type(exception).__name__))
        return await self._do_handle_failure(job, step)

    async def handle_failure(self, job: Job, step: Step, command_output: CommandOutput) -> CommandOutput:
        logger.info("Handling command failure for job {job}".format(job=job.name))
        return await self._do_handle_failure(job, step)

    async def replay_job(self, replay_request: ReplayRequest) -> ReplayResponse:
        sender_job = replay_request.sender
        target_job = replay_request.target
        if target_job not in self.wait_queues:
            self.wait_queues[target_job] = Condition()
        wait_queue = self.wait_queues[target_job]
        async with wait_queue:
            if (target_job not in self.replay_cache or
                    self.replay_cache[target_job].version < replay_request.version):
                # Reschedule job
                logger.info("Rescheduling job {job}".format(job=target_job))
                command_output = CommandOutput(value=None, status=Status.FAILED)
                self.replay_cache[target_job] = ReplayResponse(
                    job=target_job,
                    outputs=None,
                    version=self.jobs[target_job].version + 1)
                try:
                    await self.context.scheduler.notify_status(sender_job, Status.WAITING)
                    command_output = await self._replay_job(self.jobs[target_job])
                finally:
                    await self.context.scheduler.notify_status(target_job, command_output.status)
                # Retrieve output
                output_ports = target_job.step.output_ports
                output_tasks = []
                for output_port in output_ports:
                    output_tasks.append(asyncio.create_task(
                        target_job.step.output_token_processors[output_port].compute_token(target_job, command_output)))
                self.replay_cache[target_job].outputs = {port.name: token for (port, token) in
                                                         zip(output_ports, await asyncio.gather(*output_tasks))}
                wait_queue.notify_all()
            elif self.replay_cache[target_job].outputs is None:
                # Wait for job completion
                await wait_queue.wait()
            return self.replay_cache[target_job]


class DummyFailureManager(FailureManager):

    async def close(self):
        ...

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'dummy_failure_manager.json'))

    async def handle_exception(self,
                               job: Job,
                               step: Step,
                               exception: BaseException) -> CommandOutput:
        raise exception

    async def handle_failure(self,
                             job: Job,
                             step: Step,
                             command_output: CommandOutput) -> CommandOutput:
        return command_output



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pkg_resources
from jsonref import loads
from jsonschema import Draft7Validator
from ruamel.yaml import YAML

from streamflow.core import utils
from streamflow.core.exception import WorkflowDefinitionException
from streamflow.data import data_manager_classes
from streamflow.deployment import deployment_manager_classes
from streamflow.deployment.connector import connector_classes
from streamflow.persistence import database_classes
from streamflow.recovery import checkpoint_manager_classes, failure_manager_classes
from streamflow.scheduling.policy import policy_classes

if TYPE_CHECKING:
    from typing import Any, MutableMapping


def handle_errors(errors):
    errors = list(sorted(errors, key=str))
    if not errors:
        return
    raise WorkflowDefinitionException(
        "The StreamFlow configuration is invalid because:\n{error_msgs}".format(
            error_msgs="\n".join([" - {msg}".format(msg=err) for err in errors])))


def load_jsonschema(config_file: MutableMapping[str, Any]):
    try:
        base_path = pkg_resources.resource_filename(
            __name__, os.path.join('schemas', config_file['version']))
    except pkg_resources.ResolutionError:
        raise Exception(
            'Version {} is unsupported'.format(config_file['version']))
    filename = os.path.join(base_path, "config_schema.json")
    if not os.path.exists(filename):
        raise Exception(
            'Version in "{}" is unsupported'.format(filename))
    with open(filename, "r") as f:
        return loads(f.read(), base_uri='file://{}/'.format(base_path), jsonschema=True)


class SfValidator(object):

    def __init__(self) -> None:
        super().__init__()
        self.yaml = YAML(typ='safe')

    def validate_file(self, streamflow_file: str) -> MutableMapping[str, Any]:
        with open(streamflow_file) as f:
            streamflow_config = self.yaml.load(f)
        return self.validate(streamflow_config)

    def validate(self, streamflow_config: MutableMapping[str, Any]):
        schema = load_jsonschema(streamflow_config)
        utils.inject_schema(schema, checkpoint_manager_classes, 'checkpointManager')
        utils.inject_schema(schema, database_classes, 'database')
        utils.inject_schema(schema, data_manager_classes, 'dataManager')
        utils.inject_schema(schema, connector_classes, 'deployment')
        utils.inject_schema(schema, deployment_manager_classes, 'deploymentManager')
        utils.inject_schema(schema, failure_manager_classes, 'failureManager')
        utils.inject_schema(schema, policy_classes, 'policy')
        validator = Draft7Validator(schema)
        handle_errors(validator.iter_errors(streamflow_config))
        return streamflow_config



from __future__ import annotations

import asyncio
import json
import sys
import uuid
from abc import ABC, abstractmethod
from asyncio import Queue
from enum import Enum
from typing import MutableMapping, MutableSequence, TYPE_CHECKING, Type, TypeVar, cast

from streamflow.core import utils
from streamflow.core.exception import WorkflowExecutionException
from streamflow.core.persistence import DatabaseLoadingContext, DependencyType, PersistableEntity

if TYPE_CHECKING:
    from streamflow.core.deployment import Connector, Target
    from streamflow.core.context import StreamFlowContext
    from typing import Optional, Any


class Command(ABC):

    def __init__(self, step: Step):
        super().__init__()
        self.step: Step = step

    @abstractmethod
    async def execute(self, job: Job) -> CommandOutput:
        ...


class CommandOutput(object):
    __slots__ = ('value', 'status')

    def __init__(self,
                 value: Any,
                 status: Status):
        self.value: Any = value
        self.status: Status = status

    def update(self, value: Any):
        return CommandOutput(value=value, status=self.status)


class CommandOutputProcessor(ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 target: Optional[Target] = None):
        self.name: str = name
        self.workflow: Workflow = workflow
        self.target: Optional[Target] = target

    def _get_connector(self, connector: Optional[Connector], job: Job) -> Connector:
        return connector or self.workflow.context.scheduler.get_connector(job.name)

    async def _get_locations(self, connector: Optional[Connector], job: Job) -> MutableSequence[str]:
        if self.target:
            return list((await connector.get_available_locations(
                service=self.target.service)).keys())
        else:
            return self.workflow.context.scheduler.get_locations(job.name)

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CommandOutputProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            target=(await loading_context.load_target(context, row['workflow'])) if row['target'] else None)

    async def _save_additional_params(self, context: StreamFlowContext):
        if self.target:
            await self.target.save(context)
        return {'name': self.name,
                'workflow': self.workflow.persistent_id,
                'target': self.target.persistent_id if self.target else None}

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   row: MutableMapping[str, Any],
                   loading_context: DatabaseLoadingContext) -> CommandOutputProcessor:
        type = cast(Type[CommandOutputProcessor], utils.get_class_from_name(row['type']))
        return await type._load(context, row['params'], loading_context)

    @abstractmethod
    async def process(self,
                      job: Job,
                      command_output: CommandOutput,
                      connector: Optional[Connector] = None) -> Optional[Token]:
        ...

    async def save(self, context: StreamFlowContext):
        return {'type': utils.get_class_fullname(type(self)),
                'params': await self._save_additional_params(context)}


class Executor(ABC):

    def __init__(self, workflow: Workflow):
        self.workflow: Workflow = workflow

    @abstractmethod
    async def run(self) -> MutableMapping[str, Any]:
        ...


class Job(object):
    __slots__ = ('name', 'inputs', 'input_directory', 'output_directory', 'tmp_directory')

    def __init__(self,
                 name: str,
                 inputs: MutableMapping[str, Token],
                 input_directory: str,
                 output_directory: str,
                 tmp_directory: str):
        self.name: str = name
        self.inputs: MutableMapping[str, Token] = inputs
        self.input_directory: str = input_directory
        self.output_directory: str = output_directory
        self.tmp_directory: str = tmp_directory


class Port(PersistableEntity):

    def __init__(self,
                 workflow: Workflow,
                 name: str):
        super().__init__()
        self.queues: MutableMapping[str, Queue] = {}
        self.name: str = name
        self.token_list: MutableSequence[Token] = []
        self.workflow: Workflow = workflow

    def _init_consumer(self, consumer: str):
        self.queues[consumer] = Queue()
        for t in self.token_list:
            self.queues[consumer].put_nowait(t)

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> Port:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {}

    def close(self, consumer: str):
        if consumer in self.queues:
            self.queues[consumer].task_done()

    def empty(self) -> bool:
        return not self.token_list

    async def get(self, consumer: str) -> Token:
        if consumer not in self.queues:
            self._init_consumer(consumer)
            return await self.queues[consumer].get()
        else:
            token = await self.queues[consumer].get()
            self.queues[consumer].task_done()
            return token

    def get_input_steps(self) -> MutableSequence[Step]:
        return [s for s in self.workflow.steps.values() if self.name in s.output_ports.values()]

    def get_output_steps(self) -> MutableSequence[Step]:
        return [s for s in self.workflow.steps.values() if self.name in s.input_ports.values()]

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> Port:
        row = await context.database.get_port(persistent_id)
        type = cast(Type[Port], utils.get_class_from_name(row['type']))
        port = await type._load(context, row, loading_context)
        port.persistent_id = persistent_id
        loading_context.add_port(persistent_id, port)
        return port

    def put(self, token: Token):
        self.token_list.append(token)
        for q in self.queues.values():
            q.put_nowait(token)

    async def save(self, context: StreamFlowContext) -> None:
        async with self.persistence_lock:
            if not self.persistent_id:
                self.persistent_id = await context.database.add_port(
                    name=self.name,
                    workflow_id=self.workflow.persistent_id,
                    type=type(self),
                    params=json.dumps(await self._save_additional_params(context)))


class Status(Enum):
    WAITING = 0
    FIREABLE = 1
    RUNNING = 2
    SKIPPED = 3
    COMPLETED = 4
    FAILED = 5
    CANCELLED = 6


class Step(PersistableEntity, ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__()
        self.input_ports: MutableMapping[str, str] = {}
        self.name: str = name
        self.output_ports: MutableMapping[str, str] = {}
        self.status: Status = Status.WAITING
        self.terminated: bool = False
        self.workflow: Workflow = workflow

    def _add_port(self,
                  name: str,
                  port: Port,
                  type: DependencyType):
        if port.name not in self.workflow.ports:
            self.workflow.ports[port.name] = port
        if type == DependencyType.INPUT:
            self.input_ports[name] = port.name
        else:
            self.output_ports[name] = port.name

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext):
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {}

    async def _set_status(self, status: Status):
        self.status = status
        if self.persistent_id is not None:
            await self.workflow.context.database.update_step(self.persistent_id, {"status": status.value})

    def add_input_port(self, name: str, port: Port) -> None:
        self._add_port(name, port, DependencyType.INPUT)

    def add_output_port(self, name: str, port: Port) -> None:
        self._add_port(name, port, DependencyType.OUTPUT)

    def get_input_port(self, name: Optional[str] = None) -> Port:
        if name is None:
            if len(self.input_ports) == 1:
                return self.workflow.ports.get(next(iter(self.input_ports.values())))
            else:
                raise WorkflowExecutionException(
                    "Cannot retrieve default input port as step {step} contains multiple input ports.".format(
                        step=self.name))
        return self.workflow.ports.get(self.input_ports[name]) if name in self.input_ports else None

    def get_input_ports(self) -> MutableMapping[str, Port]:
        return {k: self.workflow.ports[v] for k, v in self.input_ports.items()}

    def get_output_port(self, name: Optional[str] = None) -> Port:
        if name is None:
            if len(self.output_ports) == 1:
                return self.workflow.ports.get(next(iter(self.output_ports.values())))
            else:
                raise WorkflowExecutionException(
                    "Cannot retrieve default output port as step {step} contains multiple output ports.".format(
                        step=self.name))
        return self.workflow.ports.get(self.output_ports[name]) if name in self.output_ports else None

    def get_output_ports(self) -> MutableMapping[str, Port]:
        return {k: self.workflow.ports[v] for k, v in self.output_ports.items()}

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> Step:
        row = await context.database.get_step(persistent_id)
        type = cast(Type[Step], utils.get_class_from_name(row['type']))
        step = await type._load(context, row, loading_context)
        step.persistent_id = persistent_id
        step.status = Status(row['status'])
        step.terminated = step.status in [Status.COMPLETED, Status.FAILED, Status.SKIPPED]
        ports = await context.database.get_input_ports(persistent_id)
        step.input_ports = {k: v for k, v in zip(
            [p['name'] for p in ports],
            await asyncio.gather(*(asyncio.create_task(loading_context.load_port(context, p['port'])) for p in ports)))}
        ports = await context.database.get_output_ports(persistent_id)
        step.output_ports = {k: v for k, v in zip(
            [p['name'] for p in ports],
            await asyncio.gather(*(asyncio.create_task(loading_context.load_port(context, p['port'])) for p in ports)))}
        loading_context.add_step(persistent_id, step)
        return step

    @abstractmethod
    async def run(self):
        ...

    async def save(self, context: StreamFlowContext) -> None:
        async with self.persistence_lock:
            if not self.persistent_id:
                self.persistent_id = await context.database.add_step(
                    name=self.name,
                    workflow_id=self.workflow.persistent_id,
                    status=cast(int, self.status.value),
                    type=type(self),
                    params=json.dumps(await self._save_additional_params(context)))
        save_tasks = []
        for name, port in self.get_input_ports().items():
            save_tasks.append(asyncio.create_task(self.workflow.context.database.add_dependency(
                step=self.persistent_id,
                port=port.persistent_id,
                type=DependencyType.INPUT,
                name=name)))
        for name, port in self.get_output_ports().items():
            save_tasks.append(asyncio.create_task(self.workflow.context.database.add_dependency(
                step=self.persistent_id,
                port=port.persistent_id,
                type=DependencyType.OUTPUT,
                name=name)))
        await asyncio.gather(*save_tasks)

    @abstractmethod
    async def terminate(self, status: Status):
        ...


class Token(PersistableEntity):
    __slots__ = ('persistent_id', 'value', 'tag')

    def __init__(self,
                 value: Any,
                 tag: str = '0'):
        super().__init__()
        self.value: Any = value
        self.tag: str = tag

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> Token:
        value = json.loads(row['value'])
        if isinstance(value, MutableMapping) and 'token' in value:
            value = await loading_context.load_token(context, value['token'])
        return cls(
            tag=row['tag'],
            value=value)

    async def _save_value(self, context: StreamFlowContext):
        return {'token': self.value.persistent_id} if isinstance(self.value, Token) else self.value

    async def get_weight(self, context: StreamFlowContext):
        if isinstance(self.value, Token):
            return await self.value.get_weight(context)
        else:
            return sys.getsizeof(self.value)

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> Token:
        row = await context.database.get_token(persistent_id)
        type = cast(Type[Token], utils.get_class_from_name(row['type']))
        token = await type._load(context, row, loading_context)
        token.persistent_id = persistent_id
        loading_context.add_token(persistent_id, token)
        return token

    async def is_available(self, context: StreamFlowContext):
        if isinstance(self.value, Token):
            return await self.value.is_available(context)
        else:
            return True

    def retag(self, tag: str) -> Token:
        return self.__class__(tag=tag, value=self.value)

    async def save(self,
                   context: StreamFlowContext,
                   port_id: Optional[int] = None):
        async with self.persistence_lock:
            if not self.persistent_id:
                self.persistent_id = await context.database.add_token(
                    port=port_id,
                    tag=self.tag,
                    type=type(self),
                    value=json.dumps(await self._save_value(context)))

    def update(self, value: Any) -> Token:
        return self.__class__(tag=self.tag, value=value)


class TokenProcessor(ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        self.name: str = name
        self.workflow: Workflow = workflow

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> TokenProcessor:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']))

    async def _save_additional_params(self, context: StreamFlowContext):
        return {
            'name': self.name,
            'workflow': self.workflow.persistent_id}

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   row: MutableMapping[str, Any],
                   loading_context: DatabaseLoadingContext):
        type = cast(Type[TokenProcessor], utils.get_class_from_name(row['type']))
        return await type._load(context, row['params'], loading_context)

    @abstractmethod
    async def process(self, inputs: MutableMapping[str, Token], token: Token) -> Token:
        ...

    async def save(self, context: StreamFlowContext):
        return {'type': utils.get_class_fullname(type(self)),
                'params': await self._save_additional_params(context)}


if TYPE_CHECKING:
    J = TypeVar('J', bound=Job)
    P = TypeVar('P', bound=Port)
    S = TypeVar('S', bound=Step)


class Workflow(PersistableEntity):

    def __init__(self,
                 context: StreamFlowContext,
                 type: str,
                 name: str = str(uuid.uuid4())):
        super().__init__()
        self.context: StreamFlowContext = context
        self.type: str = type
        self.name: str = name
        self.output_ports: MutableMapping[str, str] = {}
        self.ports: MutableMapping[str, Port] = {}
        self.steps: MutableMapping[str, Step] = {}

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {}

    def create_port(self,
                    cls: Type[P] = Port,
                    name: str = None,
                    **kwargs) -> P:
        if name is None:
            name = str(uuid.uuid4())
        port = cls(workflow=self, name=name, **kwargs)
        self.ports[name] = port
        return port

    def create_step(self,
                    cls: Type[S],
                    name: str = None,
                    **kwargs) -> S:
        if name is None:
            name = str(uuid.uuid4())
        step = cls(name=name, workflow=self, **kwargs)
        self.steps[name] = step
        return step

    def get_output_port(self, name: str) -> Port:
        return self.ports[self.output_ports[name]]

    def get_output_ports(self) -> MutableMapping[str, Port]:
        return {name: self.ports[p] for name, p in self.output_ports.items()}

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> Workflow:
        row = await context.database.get_workflow(persistent_id)
        workflow = cls(
            context=context,
            type=row['type'],
            name=row['name'])
        workflow.persistent_id = row['id']
        loading_context.add_workflow(persistent_id, workflow)
        rows = await context.database.get_workflow_ports(persistent_id)
        workflow.ports = {r['name']: v for r, v in zip(
            rows,
            await asyncio.gather(*(asyncio.create_task(
                loading_context.load_port(context, row['id'])) for row in rows)))}
        rows = await context.database.get_workflow_steps(persistent_id)
        workflow.steps = {r['name']: v for r, v in zip(
            rows,
            await asyncio.gather(*(asyncio.create_task(
                loading_context.load_step(context, row['id'])) for row in rows)))}
        return workflow

    async def save(self, context: StreamFlowContext) -> None:
        async with self.persistence_lock:
            if not self.persistent_id:
                self.persistent_id = await self.context.database.add_workflow(
                    name=self.name,
                    params=json.dumps(await self._save_additional_params(context)),
                    status=Status.WAITING.value,
                    type=self.type)
        await asyncio.gather(*(asyncio.create_task(port.save(context)) for port in self.ports.values()))
        await asyncio.gather(*(asyncio.create_task(step.save(context)) for step in self.steps.values()))



import argparse
import csv
import datetime

from streamflow.core.context import StreamFlowContext


def _export_to_file(fig, args: argparse.Namespace, default_name: str) -> None:
    import plotly.io as pio
    if args.format == "html":
        pio.write_html(fig, file=args.name or default_name + ".html")
    elif args.format == "json":
        pio.write_json(fig, file=args.name or default_name + ".json")
    else:
        pio.write_image(fig, format=args.format, file=args.name or "{}.{}".format(default_name, args.format))


def _timestamp_to_datetime(timestamp):
    dt = datetime.datetime.fromtimestamp(timestamp / 1e9)
    return '{}{:03.0f}'.format(dt.strftime('%Y-%m-%dT%H:%M:%S.%f'), timestamp % 1e3)


async def create_report(context: StreamFlowContext,
                        args: argparse.Namespace):
    import plotly.express as px
    # Retrieve data
    report = await context.database.get_report(args.workflow)
    # If output format is csv, print DataFrame and exit
    if args.format == 'csv':
        with open(args.name or "report.csv", 'w') as f:
            writer = csv.DictWriter(f, report[0].keys())
            writer.writeheader()
            writer.writerows(report)
            return
    # Pre-process data
    import pandas as pd
    df = pd.DataFrame(data=report)
    df["id"] = df["id"].map(str)
    df["start_time"] = pd.to_datetime(df["start_time"])
    df["end_time"] = pd.to_datetime(df["end_time"])
    # Create chart
    fig = px.timeline(
        df,
        x_start="start_time",
        x_end="end_time",
        y="name" if args.group_by_step else "id",
        color="name")
    fig.update_yaxes(visible=False)
    # Export to file
    _export_to_file(fig, args, "report")



from __future__ import annotations

import asyncio
import os
import tempfile
from typing import TYPE_CHECKING

import pkg_resources

from streamflow.core import utils
from streamflow.core.data import DataLocation, LOCAL_LOCATION
from streamflow.core.recovery import CheckpointManager
from streamflow.core.utils import random_name

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from typing import Optional, MutableSequence


class DefaultCheckpointManager(CheckpointManager):

    def __init__(self,
                 context: StreamFlowContext,
                 checkpoint_dir: Optional[str] = None):
        super().__init__(context)
        self.checkpoint_dir = checkpoint_dir or os.path.join(
            tempfile.gettempdir(), 'streamflow', 'checkpoint', utils.random_name())
        self.copy_tasks: MutableSequence = []

    async def _async_local_copy(self, data_location: DataLocation):
        parent_directory = os.path.join(self.checkpoint_dir, random_name())
        local_path = os.path.join(parent_directory, data_location.relpath)
        await self.context.data_manager.transfer_data(
            src_deployment=data_location.deployment,
            src_locations=[data_location.location],
            src_path=data_location.path,
            dst_deployment=LOCAL_LOCATION,
            dst_locations=[LOCAL_LOCATION],
            dst_path=local_path)

    async def close(self):
        pass

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'default_checkpoint_manager.json'))

    def register(self, data_location: DataLocation) -> None:
        self.copy_tasks.append(asyncio.create_task(
            self._async_local_copy(data_location)))


class DummyCheckpointManager(CheckpointManager):

    async def close(self):
        pass

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'dummy_checkpoint_manager.json'))

    def register(self, data_location: DataLocation) -> None:
        pass



import json
from typing import Any, MutableMapping, MutableSequence, Optional

from streamflow.core.context import StreamFlowContext
from streamflow.core.exception import WorkflowDefinitionException, WorkflowExecutionException
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.utils import get_tag
from streamflow.core.workflow import Port, Token, TokenProcessor, Workflow
from streamflow.cwl import utils
from streamflow.workflow.token import ListToken
from streamflow.workflow.transformer import ManyToOneTransformer, OneToOneTransformer


class AllNonNullTransformer(OneToOneTransformer):

    def _transform(self, name: str, token: Token) -> Token:
        if isinstance(token, ListToken):
            return token.update([t for t in token.value if utils.get_token_value(t) is not None])
        elif isinstance(token.value, Token):
            return token.update(self._transform(name, token.value))
        else:
            raise WorkflowExecutionException("Invalid value for token {}".format(name))

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        return {self.get_output_name(): self._transform(*next(iter(inputs.items())))}


class CWLDefaultTransformer(ManyToOneTransformer):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 default_port: Port):
        super().__init__(name, workflow)
        self.default_port: Port = default_port
        self.default_token: Optional[Token] = None

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'default_port': self.default_port.persistent_id}}

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        if len(inputs) != 1:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))
        if not self.default_port:
            raise WorkflowDefinitionException("{} step must contain a default port.".format(self.name))
        primary_token = next(iter(inputs[k] for k in inputs))
        if utils.get_token_value(primary_token) is not None:
            return {self.get_output_name(): primary_token}
        else:
            if not self.default_token:
                self.default_token = (await self._get_inputs({'__default__': self.default_port}))['__default__']
            return {self.get_output_name(): self.default_token.retag(primary_token.tag)}


class CWLTokenTransformer(ManyToOneTransformer):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 port_name: str,
                 processor: TokenProcessor):
        super().__init__(name, workflow)
        self.port_name: str = port_name
        self.processor: TokenProcessor = processor

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext):
        params = json.loads(row['params'])
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            port_name=params['port_name'],
            processor=await TokenProcessor.load(context, params['processor'], loading_context))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'port_name': self.port_name,
                   'processor': await self.processor.save(context)}}

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        return {self.get_output_name(): await self.processor.process(inputs, inputs[self.port_name])}


class FirstNonNullTransformer(OneToOneTransformer):

    def _transform(self, name: str, token: Token) -> Token:
        if isinstance(token, ListToken):
            for t in token.value:
                if utils.get_token_value(t) is not None:
                    return t
            raise WorkflowExecutionException("All sources are null in token {}".format(name))
        elif isinstance(token.value, Token):
            return token.update(self._transform(name, token.value))
        else:
            raise WorkflowExecutionException("Invalid value for token {}".format(name))

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        return {self.get_output_name(): self._transform(*next(iter(inputs.items())))}


class ForwardTransformer(OneToOneTransformer):

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        return {self.get_output_name(): next(iter(inputs.values()))}


class ListToElementTransformer(OneToOneTransformer):

    def _transform(self, token: Token) -> Token:
        if isinstance(token, ListToken):
            if len(token.value) == 1:
                return token.value[0]
            else:
                return token.update(token.value)
        elif isinstance(token.value, Token):
            return token.update(self._transform(token.value))
        else:
            return token.value

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        return {self.get_output_name(): self._transform(next(iter(inputs.values())))}


class OnlyNonNullTransformer(OneToOneTransformer):

    def _transform(self, name: str, token: Token):
        if isinstance(token, ListToken):
            ret = None
            for t in token.value:
                if utils.get_token_value(t) is not None:
                    if ret is not None:
                        raise WorkflowExecutionException(
                            "Expected only one source to be non-null in token {}".format(name))
                    ret = t
            if ret is None:
                raise WorkflowExecutionException("All sources are null in token {}".format(name))
            return ret
        elif isinstance(token.value, Token):
            return token.update(self._transform(name, token.value))
        else:
            raise WorkflowExecutionException("Invalid value for token {}".format(name))

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        return {self.get_output_name(): self._transform(*next(iter(inputs.items())))}


class ValueFromTransformer(ManyToOneTransformer):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 port_name: str,
                 processor: TokenProcessor,
                 value_from: str,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 full_js: bool = False):
        super().__init__(name, workflow)
        self.port_name: str = port_name
        self.processor: TokenProcessor = processor
        self.value_from: str = value_from
        self.expression_lib: Optional[MutableSequence[str]] = expression_lib
        self.full_js: bool = full_js

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'port_name': self.port_name,
                   'processor': await self.processor.save(context),
                   'value_from': self.value_from,
                   'expression_lib': self.expression_lib,
                   'full_js': self.full_js}}

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        if self.get_output_name() in inputs:
            inputs = {
                **inputs,
                **{self.get_output_name(): await self.processor.process(inputs, inputs[self.get_output_name()])}}
        context = utils.build_context(inputs)
        context = {**context, **{'self': context['inputs'].get(self.get_output_name())}}
        return {self.get_output_name(): Token(
            tag=get_tag(inputs.values()),
            value=utils.eval_expression(
                expression=self.value_from,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib))}


class LoopValueFromTransformer(ValueFromTransformer):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 port_name: str,
                 processor: TokenProcessor,
                 value_from: str,
                 expression_lib: Optional[MutableSequence[str]] = None,
                 full_js: bool = False):
        super().__init__(name, workflow, port_name, processor, value_from, expression_lib, full_js)
        self.loop_input_ports: MutableSequence[str] = []
        self.loop_source_port: Optional[str] = None

    def add_loop_input_port(self, name: str, port: Port):
        self.add_input_port(name + "-in", port)
        self.loop_input_ports.append(name)

    def add_loop_source_port(self, name: str, port: Port):
        self.add_input_port(name + "-out", port)
        self.loop_source_port = name

    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        loop_inputs = {k: inputs[k + "-in"] for k in self.loop_input_ports}
        self_token = (await self.processor.process(loop_inputs, inputs[self.loop_source_port + "-out"])
                      if self.loop_source_port else None)
        context = utils.build_context(loop_inputs)
        context = {**context, **{'self': utils.get_token_value(self_token)}}
        return {self.get_output_name(): Token(
            tag=get_tag(inputs.values()),
            value=utils.eval_expression(
                expression=self.value_from,
                context=context,
                full_js=self.full_js,
                expression_lib=self.expression_lib))}



import argparse
import json
import logging
import os

import cwltool.context
import cwltool.load_tool
import cwltool.loghandler
import cwltool.main
import cwltool.utils
from cwltool.resolver import tool_resolver

from streamflow.config.config import WorkflowConfig
from streamflow.core.context import StreamFlowContext
from streamflow.cwl.translator import CWLTranslator
from streamflow.log_handler import logger
from streamflow.workflow.executor import StreamFlowExecutor


def _parse_arg(path: str, context: StreamFlowContext):
    if '://' in path:
        return [path]
    elif os.path.isabs(path):
        return [os.path.join(context.config_dir, path)]
    else:
        return [path]


def _parse_args(workflow_config: WorkflowConfig, context: StreamFlowContext):
    cwl_config = workflow_config.config
    args = _parse_arg(cwl_config['file'], context)
    if 'settings' in cwl_config:
        args.extend(_parse_arg(cwl_config['settings'], context))
    return args


async def main(workflow_config: WorkflowConfig, context: StreamFlowContext, args: argparse.Namespace):
    # Parse input arguments
    cwl_args = _parse_args(workflow_config, context)
    # Configure log level
    if args.quiet:
        # noinspection PyProtectedMember
        cwltool.loghandler._logger.setLevel(logging.WARN)
    # Load CWL workflow definition
    loading_context = cwltool.context.LoadingContext()
    loading_context.resolver = tool_resolver
    loading_context.loader = cwltool.load_tool.default_loader(
        loading_context.fetcher_constructor)
    loading_context, workflowobj, uri = cwltool.load_tool.fetch_document(cwl_args[0], loading_context)
    cwltool.main.setup_schema(argparse.Namespace(enable_ext=True), None)
    loading_context, uri = cwltool.load_tool.resolve_and_validate_document(
        loading_context, workflowobj, uri
    )
    cwl_definition = cwltool.load_tool.make_tool(uri, loading_context)
    if len(cwl_args) == 2:
        loader = cwltool.load_tool.default_loader(
            loading_context.fetcher_constructor)
        loader.add_namespaces(cwl_definition.metadata.get('$namespaces', {}))
        cwl_inputs, _ = loader.resolve_ref(
            cwl_args[1],
            checklinks=False,
            content_types=cwltool.CWL_CONTENT_TYPES)

        def expand_formats(p) -> None:
            if "format" in p:
                p["format"] = loader.expand_url(p["format"], "")

        cwltool.utils.visit_class(cwl_inputs, ("File",), expand_formats)
    else:
        cwl_inputs = {}
    # Transpile CWL workflow to the StreamFlow representation
    logger.info("Processing workflow {}".format(args.name))
    translator = CWLTranslator(
        context=context,
        name=args.name,
        output_directory=args.outdir,
        cwl_definition=cwl_definition,
        cwl_inputs=cwl_inputs,
        workflow_config=workflow_config,
        loading_context=loading_context)
    logger.info("Building workflow execution plan")
    workflow = translator.translate()
    await workflow.save(context)
    logger.info("Building of workflow execution plan terminated with status COMPLETED")
    executor = StreamFlowExecutor(workflow)
    logger.info("Running workflow {}".format(args.name))
    output_tokens = await executor.run()
    logger.info("Workflow execution terminated with status COMPLETED")
    print(json.dumps(output_tokens, sort_keys=True, indent=4))



import argparse
import asyncio
import logging
import os
import sys
from typing import Any, MutableMapping, Optional, Type

from streamflow import report
from streamflow.config.config import WorkflowConfig
from streamflow.config.validator import SfValidator
from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.exception import WorkflowException
from streamflow.cwl.main import main as cwl_main
from streamflow.data import data_manager_classes
from streamflow.deployment import deployment_manager_classes
from streamflow.log_handler import logger
from streamflow.parser import parser
from streamflow.persistence import database_classes
from streamflow.recovery import checkpoint_manager_classes, failure_manager_classes
from streamflow.scheduling import scheduler_classes


async def _async_list(args: argparse.Namespace):
    if os.path.exists(args.streamflow_file):
        streamflow_config = SfValidator().validate_file(args.streamflow_file)
        context = build_context(os.path.dirname(args.streamflow_file), streamflow_config, args.outdir)
    else:
        context = build_context(os.getcwd(), {}, args.outdir)
    try:
        workflows = await context.database.list_workflows()
        if workflows:
            max_sizes = {k: len(max(w[k] for w in workflows)) for k in workflows[0].keys()}
            format_string = ("{:<" + str(max(max_sizes['name'], 4)) + "} " +
                             "{:<" + str(max(max_sizes['type'], 4)) + "} " +
                             "{:<" + str(max(max_sizes['status'], 6)) + "}")
            print(format_string.format('NAME', 'TYPE', 'STATUS'))
            for w in workflows:
                print(format_string.format(w['name'], w['type'], w['status']))
        else:
            print("No workflow objects found.")
    finally:
        await context.close()


async def _async_main(args: argparse.Namespace):
    utils.load_extensions()
    streamflow_config = SfValidator().validate_file(args.streamflow_file)
    context = build_context(os.path.dirname(args.streamflow_file), streamflow_config, args.outdir)
    try:
        workflow_tasks = []
        for workflow in streamflow_config.get('workflows', {}):
            workflow_config = WorkflowConfig(workflow, streamflow_config)
            if workflow_config.type == 'cwl':
                workflow_tasks.append(asyncio.create_task(cwl_main(workflow_config, context, args)))
            await asyncio.gather(*workflow_tasks)
    except BaseException as e:
        logger.exception(e)
        sys.exit(1)
    finally:
        await context.close()


async def _async_report(args: argparse.Namespace):
    if os.path.exists(args.streamflow_file):
        streamflow_config = SfValidator().validate_file(args.streamflow_file)
        context = build_context(os.path.dirname(args.streamflow_file), streamflow_config, args.outdir)
    else:
        context = build_context(os.getcwd(), {}, args.outdir)
    try:
        await report.create_report(context, args)
    finally:
        await context.close()


def _get_instance_from_config(
        streamflow_config: MutableMapping[str, Any],
        classes: MutableMapping[str, Type],
        instance_type: str,
        kwargs: MutableMapping[str, Any],
        enabled_by_default: bool = True) -> Any:
    config = streamflow_config.get(instance_type, None)
    if config is not None:
        enabled = config.get('enabled', enabled_by_default)
        class_name = config.get('type', 'default' if enabled else 'dummy')
        kwargs = {**kwargs, **config.get('config', {})}
    else:
        class_name = 'default' if enabled_by_default else 'dummy'
    class_ = classes[class_name]
    return class_(**kwargs)


def build_context(config_dir: str,
                  streamflow_config: MutableMapping[str, Any],
                  output_dir: Optional[str] = os.getcwd()) -> StreamFlowContext:
    context = StreamFlowContext(config_dir)
    context.checkpoint_manager = _get_instance_from_config(
        streamflow_config, checkpoint_manager_classes, 'checkpointManager',
        {'context': context}, enabled_by_default=False)
    context.database = _get_instance_from_config(
        streamflow_config, database_classes, 'database',
        {'context': context, 'connection': os.path.join(output_dir, '.streamflow', 'sqlite.db')})
    context.data_manager = _get_instance_from_config(
        streamflow_config, data_manager_classes, 'dataManager', {'context': context})
    context.deployment_manager = _get_instance_from_config(
        streamflow_config, deployment_manager_classes, 'deploymentManager', {'context': context})
    context.failure_manager = _get_instance_from_config(
        streamflow_config, failure_manager_classes, 'failureManager',
        {'context': context}, enabled_by_default=False)
    context.scheduler = _get_instance_from_config(
        streamflow_config.get('scheduling', {}), scheduler_classes, 'scheduler', {
            'context': context})
    return context


def main(args):
    args = parser.parse_args(args)
    if args.context == "version":
        from streamflow.version import VERSION
        print("StreamFlow version {version}".format(version=VERSION))
    elif args.context == "list":
        asyncio.run(_async_list(args))
    elif args.context == "report":
        asyncio.run(_async_report(args))
    elif args.context == "run":
        if args.quiet:
            logger.setLevel(logging.WARN)
        elif args.debug:
            logger.setLevel(logging.DEBUG)
        asyncio.run(_async_main(args))
    else:
        raise Exception


def run():
    main(sys.argv[1:])


if __name__ == "__main__":
    main(sys.argv[1:])



from __future__ import annotations

import asyncio
import json
import os
import posixpath
import tempfile
from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING

from streamflow.core.config import Config, SchemaEntity
from streamflow.core.data import LOCAL_LOCATION
from streamflow.core.persistence import DatabaseLoadingContext, PersistableEntity

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from streamflow.core.scheduling import Location
    from typing import MutableSequence, MutableMapping, Optional, Any, Tuple, Union


def _init_workdir(deployment_name: str) -> str:
    if deployment_name != LOCAL_LOCATION:
        return posixpath.join('/tmp', 'streamflow')
    else:
        return os.path.join(tempfile.gettempdir(), 'streamflow')


class Connector(SchemaEntity):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext):
        self.deployment_name: str = deployment_name
        self.context: StreamFlowContext = context

    @abstractmethod
    async def copy(self,
                   src: str,
                   dst: str,
                   locations: MutableSequence[str],
                   kind: ConnectorCopyKind,
                   source_connector: Optional[Connector] = None,
                   source_location: Optional[str] = None,
                   read_only: bool = False) -> None:
        ...

    @abstractmethod
    async def deploy(self, external: bool) -> None:
        ...

    @abstractmethod
    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        ...

    @abstractmethod
    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        ...

    @abstractmethod
    async def undeploy(self, external: bool) -> None:
        ...


class ConnectorCopyKind(Enum):
    LOCAL_TO_REMOTE = 1
    REMOTE_TO_LOCAL = 2
    REMOTE_TO_REMOTE = 3


class DeploymentManager(SchemaEntity):

    def __init__(self,
                 context: StreamFlowContext) -> None:
        self.context: StreamFlowContext = context

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    async def deploy(self, deployment_config: DeploymentConfig):
        ...

    @abstractmethod
    def get_connector(self, deployment_name: str) -> Optional[Connector]:
        ...

    @abstractmethod
    def is_deployed(self, deployment_name: str):
        ...

    @abstractmethod
    async def undeploy(self, deployment_name: str):
        ...

    @abstractmethod
    async def undeploy_all(self):
        ...


class DeploymentConfig(Config, PersistableEntity):
    __slots__ = ('name', 'type', 'config', 'external', 'lazy', 'workdir')

    def __init__(self,
                 name: str,
                 type: str,
                 config: MutableMapping[str, Any],
                 external: bool = False,
                 lazy: bool = True,
                 workdir: Optional[str] = None) -> None:
        super().__init__(name, type, config)
        self.external = external
        self.lazy: bool = lazy
        self.workdir: Optional[str] = workdir

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> DeploymentConfig:
        row = await context.database.get_deployment(persistent_id)
        obj = cls(
            name=row['name'],
            type=row['type'],
            config=row['config'],
            external=row['external'],
            lazy=row['lazy'],
            workdir=row['workdir'])
        obj.persistent_id = persistent_id
        loading_context.add_deployment(persistent_id, obj)
        return obj

    async def save(self, context: StreamFlowContext) -> None:
        async with self.persistence_lock:
            if not self.persistent_id:
                self.persistent_id = await context.database.add_deployment(
                    name=self.name,
                    type=self.type,
                    config=json.dumps(self.config),
                    external=self.external,
                    lazy=self.lazy,
                    workdir=self.workdir)


class Target(PersistableEntity):

    def __init__(self,
                 deployment: DeploymentConfig,
                 locations: int = 1,
                 service: Optional[str] = None,
                 scheduling_group: Optional[str] = None,
                 scheduling_policy: Optional[Config] = None,
                 workdir: Optional[str] = None):
        super().__init__()
        self.deployment: DeploymentConfig = deployment
        self.locations: int = locations
        self.service: Optional[str] = service
        self.scheduling_group: Optional[str] = scheduling_group
        self.scheduling_policy: Optional[Config] = (
                scheduling_policy or Config(name='__DEFAULT__', type='data_locality', config={}))
        self.workdir: str = workdir or self.deployment.workdir or _init_workdir(deployment.name)

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {}

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> Target:
        row = await context.database.get_target(persistent_id)
        obj = cls(
            deployment=await DeploymentConfig.load(context, row['deployment'], loading_context),
            locations=row['locations'],
            service=row['service'],
            workdir=row['workdir'])
        obj.persistent_id = persistent_id
        loading_context.add_target(persistent_id, obj)
        return obj

    async def save(self, context: StreamFlowContext) -> None:
        await self.deployment.save(context)
        async with self.persistence_lock:
            if not self.persistent_id:
                self.persistent_id = await context.database.add_target(
                    deployment=self.deployment.persistent_id,
                    type=type(self),
                    params=json.dumps(await self._save_additional_params(context)),
                    locations=self.locations,
                    service=self.service,
                    workdir=self.workdir)


class LocalTarget(Target):
    __deployment_config = None

    def __init__(self, workdir: Optional[str] = None):
        super().__init__(
            deployment=self._get_deployment_config(),
            locations=1,
            workdir=workdir)

    @classmethod
    def _get_deployment_config(cls):
        if not cls.__deployment_config:
            cls.__deployment_config = DeploymentConfig(
                name=LOCAL_LOCATION,
                type='local',
                config={},
                external=True,
                lazy=False)
        return cls.__deployment_config
